# ruff: noqa: F401
from .cifar import CIFAR10C, CIFAR10H, CIFAR10N, CIFAR100C, CIFAR100N
from .cub import CUB
from .imagenet import (
    ImageNetA,
    ImageNetC,
    ImageNetO,
    ImageNetR,
    TinyImageNet,
    TinyImageNetC,
)
from .mnist_c import MNISTC
from .not_mnist import NotMNIST
from .openimage_o import OpenImageO
from .uci import (
    HTRU2,
    BankMarketing,
    DOTA2Games,
    OnlineShoppers,
    SpamBase,
)
