# ruff: noqa: F401
from .bank_marketing import BankMarketingDataModule
from .dota2_games import DOTA2GamesDataModule
from .htru2 import HTRU2DataModule
from .online_shoppers import OnlineShoppersDataModule
from .spam_base import SpamBaseDataModule
from .uci_classification import UCIClassificationDataModule
