@import "theme.css";
/* sphinx-design styles for cards/tabs
*/

:root {
    --sd-color-info: #29b99a;
    --sd-color-primary: #6c6c6d;
    --sd-color-primary-highlight: #f3f4f7;
    --sd-color-card-border-hover: #29b99a;
    --sd-color-card-border: #f3f4f7;
    --sd-color-card-background: #fff;
    --sd-color-card-text: inherit;
    --sd-color-card-header: transparent;
    --sd-color-card-footer: transparent;
    --sd-color-tabs-label-active: hsla(231, 99%, 66%, 1);
    --sd-color-tabs-label-hover: hsla(231, 99%, 66%, 1);
    --sd-color-tabs-label-inactive: hsl(0, 0%, 66%);
    --sd-color-tabs-underline-active: hsla(231, 99%, 66%, 1);
    --sd-color-tabs-underline-hover: rgba(178, 206, 245, 0.62);
    --sd-color-tabs-underline-inactive: transparent;
    --sd-color-tabs-overline: rgb(222, 222, 222);
    --sd-color-tabs-underline: rgb(222, 222, 222);
}

.sd-text-info {
    color: #29b99a;
}


.sd-card {
    position: relative;
    background-color: #fff;
    opacity: 1.0;
    border-radius: 0px;
    width: 30%;
    border: none;
    padding-bottom: 0px;
}


.sd-card-img {
    opacity: 0.5;
    width: 200px;
    padding: 0px;
}

.sd-card-img:hover {
    opacity: 1.0;
    background-color: #f3f4f7;
}


.sd-card:after {
    display: block;
    opacity: 1;
    content: '';
    border-bottom: solid 1px #29b99a;
    background-color: #fff;
    transform: scaleX(0);
    transition: transform .250s ease-in-out;
    transform-origin: 0% 50%;
}

.sd-card:hover {
    background-color: #fff;
    opacity: 1;
    border-top: 1px solid #f3f4f7;
    border-left: 1px solid #f3f4f7;
    border-right: 1px solid #f3f4f7;
}

.sd-card:hover:after {
    transform: scaleX(1);
}

.header-logo {
    background-image: url("../images/torch_uncertainty.png");
    background-size: 220px 40px;
    height: 40px;
    width: 220px;
    position: absolute;
    top: 0;
    bottom: 0;
    margin: auto 0;
}

.tutorials-header .header-logo {
    background-image: url("../images/torch_uncertainty.png");
}

.figure-caption {
    text-align: center;
}

pre {
    line-height: 125%;
}

td.linenos .normal {
    color: #6e7681;
    background-color: #0d1117;
    padding-left: 5px;
    padding-right: 5px;
}

span.linenos {
    color: #6e7681;
    background-color: #0d1117;
    padding-left: 5px;
    padding-right: 5px;
}

td.linenos .special {
    color: #e6edf3;
    background-color: #6e7681;
    padding-left: 5px;
    padding-right: 5px;
}

span.linenos.special {
    color: #e6edf3;
    background-color: #6e7681;
    padding-left: 5px;
    padding-right: 5px;
}

.highlight .hll {
    background-color: #6e7681
}

.highlight {
    background: #0d1117;
    color: #e6edf3
}

.highlight .c {
    color: #8b949e;
    font-style: italic
}

/* Comment */
.highlight .err {
    color: #f85149
}

/* Error */
.highlight .esc {
    color: #e6edf3
}

/* Escape */
.highlight .g {
    color: #e6edf3
}

/* Generic */
.highlight .k {
    color: #ff7b72
}

/* Keyword */
.highlight .l {
    color: #a5d6ff
}

/* Literal */
.highlight .n {
    color: #e6edf3
}

/* Name */
.highlight .o {
    color: #ff7b72;
    font-weight: bold
}

/* Operator */
.highlight .x {
    color: #e6edf3
}

/* Other */
.highlight .p {
    color: #e6edf3
}

/* Punctuation */
.highlight .ch {
    color: #8b949e;
    font-style: italic
}

/* Comment.Hashbang */
.highlight .cm {
    color: #8b949e;
    font-style: italic
}

/* Comment.Multiline */
.highlight .cp {
    color: #8b949e;
    font-weight: bold;
    font-style: italic
}

/* Comment.Preproc */
.highlight .cpf {
    color: #8b949e;
    font-style: italic
}

/* Comment.PreprocFile */
.highlight .c1 {
    color: #8b949e;
    font-style: italic
}

/* Comment.Single */
.highlight .cs {
    color: #8b949e;
    font-weight: bold;
    font-style: italic
}

/* Comment.Special */
.highlight .gd {
    color: #ffa198;
    background-color: #490202
}

/* Generic.Deleted */
.highlight .ge {
    color: #e6edf3;
    font-style: italic
}

/* Generic.Emph */
.highlight .ges {
    color: #e6edf3;
    font-weight: bold;
    font-style: italic
}

/* Generic.EmphStrong */
.highlight .gr {
    color: #ffa198
}

/* Generic.Error */
.highlight .gh {
    color: #79c0ff;
    font-weight: bold
}

/* Generic.Heading */
.highlight .gi {
    color: #56d364;
    background-color: #0f5323
}

/* Generic.Inserted */
.highlight .go {
    color: #8b949e
}

/* Generic.Output */
.highlight .gp {
    color: #8b949e
}

/* Generic.Prompt */
.highlight .gs {
    color: #e6edf3;
    font-weight: bold
}

/* Generic.Strong */
.highlight .gu {
    color: #79c0ff
}

/* Generic.Subheading */
.highlight .gt {
    color: #ff7b72
}

/* Generic.Traceback */
.highlight .g-Underline {
    color: #e6edf3;
    text-decoration: underline
}

/* Generic.Underline */
.highlight .kc {
    color: #79c0ff
}

/* Keyword.Constant */
.highlight .kd {
    color: #ff7b72
}

/* Keyword.Declaration */
.highlight .kn {
    color: #ff7b72
}

/* Keyword.Namespace */
.highlight .kp {
    color: #79c0ff
}

/* Keyword.Pseudo */
.highlight .kr {
    color: #ff7b72
}

/* Keyword.Reserved */
.highlight .kt {
    color: #ff7b72
}

/* Keyword.Type */
.highlight .ld {
    color: #79c0ff
}

/* Literal.Date */
.highlight .m {
    color: #a5d6ff
}

/* Literal.Number */
.highlight .s {
    color: #a5d6ff
}

/* Literal.String */
.highlight .na {
    color: #e6edf3
}

/* Name.Attribute */
.highlight .nb {
    color: #e6edf3
}

/* Name.Builtin */
.highlight .nc {
    color: #f0883e;
    font-weight: bold
}

/* Name.Class */
.highlight .no {
    color: #79c0ff;
    font-weight: bold
}

/* Name.Constant */
.highlight .nd {
    color: #d2a8ff;
    font-weight: bold
}

/* Name.Decorator */
.highlight .ni {
    color: #ffa657
}

/* Name.Entity */
.highlight .ne {
    color: #f0883e;
    font-weight: bold
}

/* Name.Exception */
.highlight .nf {
    color: #d2a8ff;
    font-weight: bold
}

/* Name.Function */
.highlight .nl {
    color: #79c0ff;
    font-weight: bold
}

/* Name.Label */
.highlight .nn {
    color: #ff7b72
}

/* Name.Namespace */
.highlight .nx {
    color: #e6edf3
}

/* Name.Other */
.highlight .py {
    color: #79c0ff
}

/* Name.Property */
.highlight .nt {
    color: #7ee787
}

/* Name.Tag */
.highlight .nv {
    color: #79c0ff
}

/* Name.Variable */
.highlight .ow {
    color: #ff7b72;
    font-weight: bold
}

/* Operator.Word */
.highlight .pm {
    color: #e6edf3
}

/* Punctuation.Marker */
.highlight .w {
    color: #6e7681
}

/* Text.Whitespace */
.highlight .mb {
    color: #a5d6ff
}

/* Literal.Number.Bin */
.highlight .mf {
    color: #a5d6ff
}

/* Literal.Number.Float */
.highlight .mh {
    color: #a5d6ff
}

/* Literal.Number.Hex */
.highlight .mi {
    color: #a5d6ff
}

/* Literal.Number.Integer */
.highlight .mo {
    color: #a5d6ff
}

/* Literal.Number.Oct */
.highlight .sa {
    color: #79c0ff
}

/* Literal.String.Affix */
.highlight .sb {
    color: #a5d6ff
}

/* Literal.String.Backtick */
.highlight .sc {
    color: #a5d6ff
}

/* Literal.String.Char */
.highlight .dl {
    color: #79c0ff
}

/* Literal.String.Delimiter */
.highlight .sd {
    color: #a5d6ff
}

/* Literal.String.Doc */
.highlight .s2 {
    color: #a5d6ff
}

/* Literal.String.Double */
.highlight .se {
    color: #79c0ff
}

/* Literal.String.Escape */
.highlight .sh {
    color: #79c0ff
}

/* Literal.String.Heredoc */
.highlight .si {
    color: #a5d6ff
}

/* Literal.String.Interpol */
.highlight .sx {
    color: #a5d6ff
}

/* Literal.String.Other */
.highlight .sr {
    color: #79c0ff
}

/* Literal.String.Regex */
.highlight .s1 {
    color: #a5d6ff
}

/* Literal.String.Single */
.highlight .ss {
    color: #a5d6ff
}

/* Literal.String.Symbol */
.highlight .bp {
    color: #e6edf3
}

/* Name.Builtin.Pseudo */
.highlight .fm {
    color: #d2a8ff;
    font-weight: bold
}

/* Name.Function.Magic */
.highlight .vc {
    color: #79c0ff
}

/* Name.Variable.Class */
.highlight .vg {
    color: #79c0ff
}

/* Name.Variable.Global */
.highlight .vi {
    color: #79c0ff
}

/* Name.Variable.Instance */
.highlight .vm {
    color: #79c0ff
}

/* Name.Variable.Magic */
.highlight .il {
    color: #a5d6ff
}

/* Literal.Number.Integer.Long */

.highlight-bash pre {
    background: #0d1117;
    color: #e6edf3;
}

code.highlight.cmd.bash.docutils.literal.highlight-bash {
    background: #f3f4f7;
    color: #6c6c6d;
}
