# ruff: noqa: F401
from .classification import ClassificationRoutine
from .pixel_regression import PixelRegressionRoutine
from .regression import RegressionRoutine
from .segmentation import SegmentationRoutine
