# ruff: noqa: F401
from .aggregated_dataset import AggregatedDataset
from .fractals import Fractals
from .frost import FrostImages
from .kitti import KITTIDepth
from .muad import MUAD
from .nyu import NYUv2
